@@ -13,7 +13,7 @@
  *
  * You should have received a copy of the GNU Lesser General
  * Public License along with this library; if not, see
- * <http://www.gnu.org/licenses/>.
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -27,9 +27,11 @@ static const char *class_names[] =
   "BablTypeInteger",
   "BablTypeFloat",
   "BablSampling",
+  "BablTRC",
   "BablComponent",
   "BablModel",
   "BablFormat",
+  "BablSpace",
   "BablConversion",
   "BablConversionLinear",
   "BablConversionPlane",
@@ -63,8 +65,8 @@ babl_backtrack (void)
 {
   char buf[512];
 
-  sprintf (buf, "echo bt>/tmp/babl.gdb;"
-           "gdb -q --batch -x /tmp/babl.gdb --pid=%i 2>/dev/null", getpid ());
+  snprintf (buf, sizeof (buf), "echo bt>/tmp/babl.gdb;"
+           "gdb -q --batch -x /tmp/babl.gdb --pid=%i | grep 'in ''babl_die' -A40", getpid ());
   return system (buf);
 }
 
@@ -80,6 +82,7 @@ BablMutex *babl_format_mutex;
 #if BABL_DEBUG_MEM
 BablMutex *babl_debug_mutex;
 #endif
+BablMutex *babl_reference_mutex;
 
 void
 babl_internal_init (void)
@@ -87,6 +90,7 @@ babl_internal_init (void)
   babl_set_malloc (malloc);
   babl_set_free (free);
   babl_format_mutex = babl_mutex_new ();
+  babl_reference_mutex = babl_mutex_new ();
 #if BABL_DEBUG_MEM
   babl_debug_mutex = babl_mutex_new ();
 #endif
@@ -96,6 +100,7 @@ void
 babl_internal_destroy (void)
 {
   babl_mutex_destroy (babl_format_mutex);
+  babl_mutex_destroy (babl_reference_mutex);
 #if BABL_DEBUG_MEM
   babl_mutex_destroy (babl_debug_mutex);
 #endif
@@ -108,3 +113,17 @@ babl_get_name (const Babl *babl)
   babl_assert (BABL_IS_BABL (babl));
   return babl->instance.name;
 }
+
+const char *
+babl_get_doc (const Babl *babl)
+{
+  babl_assert (BABL_IS_BABL (babl));
+  return babl->instance.doc;
+}
+
+void babl_doc (const Babl     *babl,
+               const char     *doc)
+{
+  babl_assert (BABL_IS_BABL (babl));
+  ((Babl*)babl)->instance.doc = doc;
+}
