@@ -14,7 +14,7 @@
  *    GNU General Public License for more details.
  *
  *    You should have received a copy of the GNU General Public License
- *    along with GGGL; if not, see <http://www.gnu.org/licenses/>.
+ *    along with GGGL; if not, see <https://www.gnu.org/licenses/>.
  *
  *    Rights are granted to use this shared object in libraries covered by
  *    LGPL. (exception added, during import into babl CVS.)
@@ -40,6 +40,7 @@
  */
 
 #include <stdlib.h>
+#include <stdint.h>
 
 #include "config.h"
 #include "babl.h"
@@ -48,14 +49,12 @@
 #include "extensions/util.h"
 
 
-#define INLINE    inline
-
 /* lookup tables used in conversion */
 
 static float         table_8_F[1 << 8];
 static float         table_8g_F[1 << 8];
-static unsigned char table_F_8[1 << 16];
-static unsigned char table_F_8g[1 << 16];
+static unsigned char table_F_8[1 << 17];
+static unsigned char table_F_8g[1 << 17];
 
 static int           table_inited = 0;
 
@@ -80,68 +79,75 @@ table_init (void)
   {
     union
     {
-      float          f;
-      unsigned short s[2];
+      float    f;
+      uint32_t s;
     } u;
     u.f = 0.0;
 
-    u.s[0] = 0.0;
+    //u.s[0] = 0;
 
-    for (u.s[1] = 0; u.s[1] < 65535; u.s[1] += 1)
+    for (u.s = 0; u.s < 4294900000U; u.s += 32768)
       {
-        unsigned char c;
-        unsigned char cg;
+        int c;
+        int cg;
 
         if (u.f <= 0.0)
           {
             c  = 0;
             cg = 0;
           }
-        else if (u.f >= 1.0)
-          {
-            c  = 255;
-            cg = 255;
-          }
         else
           {
-            c  = rint (u.f * 255.0);
-            cg = rint (linear_to_gamma_2_2 (u.f) * 255.0);
+            c  = (u.f * 255.1619) + 0.5;
+            cg = (linear_to_gamma_2_2 (u.f) * 255.1619) + 0.5;
+            if (cg > 255) cg = 255;
+            if (c > 255) c = 255;
           }
 
-        table_F_8[u.s[1]]  = c;
-        table_F_8g[u.s[1]] = cg;
+        table_F_8[(u.s >> 15) & ((1 << 17)-1)]  = c;
+        table_F_8g[(u.s >> 15) & ((1 << 17)-1)] = cg;
       }
   }
+#if 0
   /* fix tables to ensure 1:1 conversions back and forth */
   if (0)
     {
       int i;
       for (i = 0; i < 256; i++)
         {
-          float           f  = table_8_F[i];
+          float           f;
           unsigned short *hi = ((unsigned short *) (void *) &f);
           unsigned short *lo = ((unsigned short *) (void *) &f);
+
+          f = table_8_F[i];
           *lo              = 0;
-          table_F_8[(*hi)] = i;
+          table_F_8[*hi] = i;
+          f  = table_8g_F[i];
+          *lo              = 0;
+          table_F_8g[*hi] = i;
         }
     }
+#endif
 }
 
 /* function to find the index in table for a float */
-static unsigned int
+static inline unsigned int
 gggl_float_to_index16 (float f)
 {
   union
   {
     float          f;
-    unsigned short s[2];
+    uint32_t       s;
   } u;
   u.f = f;
-  return u.s[1];
+  return (u.s >> 15) & ((1 << 17)-1);
 }
 
-static INLINE long
-conv_F_8 (unsigned char *src, unsigned char *dst, long samples)
+static inline void
+conv_F_8 (const Babl    *conversion,
+          unsigned char *src, 
+          unsigned char *dst, 
+          long           samples)
 {
   long n = samples;
 
@@ -154,12 +160,14 @@ conv_F_8 (unsigned char *src, unsigned char *dst, long samples)
       dst                   += 1;
       src                   += 4;
     }
-  return samples;
 }
 
 
-static INLINE long
-conv_F_8g (unsigned char *src, unsigned char *dst, long samples)
+static inline void
+conv_F_8g (const Babl    *conversion,
+           unsigned char *src, 
+           unsigned char *dst, 
+           long           samples)
 {
   long n = samples;
 
@@ -172,12 +180,14 @@ conv_F_8g (unsigned char *src, unsigned char *dst, long samples)
       dst                   += 1;
       src                   += 4;
     }
-  return samples;
 }
 
 
-static INLINE long
-conv_8_F (unsigned char *src, unsigned char *dst, long samples)
+static inline void __attribute__((unused))
+conv_8_F (const Babl    *conversion,
+          unsigned char *src, 
+          unsigned char *dst, 
+          long           samples)
 {
   long n = samples;
 
@@ -189,12 +199,14 @@ conv_8_F (unsigned char *src, unsigned char *dst, long samples)
       dst             += 4;
       src             += 1;
     }
-  return samples;
 }
 
 
-static INLINE long
-conv_rgbaF_rgb8 (unsigned char *src, unsigned char *dst, long samples)
+static void
+conv_rgbaF_rgb8 (const Babl    *conversion,
+                 unsigned char *src, 
+                 unsigned char *dst, 
+                 long           samples)
 {
   long n = samples;
 
@@ -217,12 +229,14 @@ conv_rgbaF_rgb8 (unsigned char *src, unsigned char *dst, long samples)
 
       src += 4;
     }
-  return samples;
 }
 
 
-static INLINE long
-conv_rgbaF_rgba8 (unsigned char *src, unsigned char *dst, long samples)
+static void __attribute__((unused))
+conv_rgbaF_rgba8 (const Babl    *conversion,
+                  unsigned char *src, 
+                  unsigned char *dst, 
+                  long           samples)
 {
   long n = samples;
 
@@ -248,23 +262,26 @@ conv_rgbaF_rgba8 (unsigned char *src, unsigned char *dst, long samples)
       src                   += 4;
       dst                   += 1;
     }
-  return samples;
 }
 
 #define conv_rgbaF_rgbP8    conv_rgbaF_rgba8
 
-static INLINE long
-conv_rgbF_rgb8 (unsigned char *src, unsigned char *dst, long samples)
+static void __attribute__((unused))
+conv_rgbF_rgb8 (const Babl    *conversion,
+                unsigned char *src, 
+                unsigned char *dst, 
+                long           samples)
 {
-  conv_F_8g (src, dst, samples * 3);
-  return samples;
+  conv_F_8g (conversion, src, dst, samples * 3);
 }
 
-static INLINE long
-conv_gaF_ga8 (unsigned char *src, unsigned char *dst, long samples)
+static void __attribute__((unused))
+conv_gaF_ga8 (const Babl    *conversion,
+              unsigned char *src, 
+              unsigned char *dst, 
+              long           samples)
 {
-  conv_F_8 (src, dst, samples * 2);
-  return samples;
+  conv_F_8 (conversion, src, dst, samples * 2);
 }
 
 #define conv_rgbAF_rgbA8    conv_rgbaF_rgba8
@@ -272,8 +289,11 @@ conv_gaF_ga8 (unsigned char *src, unsigned char *dst, long samples)
 #define conv_gAF_gA8        conv_gaF_ga8
 
 
-static INLINE long
-conv_rgba8_rgbaF (unsigned char *src, unsigned char *dst, long samples)
+static void
+conv_rgba8_rgbaF (const Babl    *conversion,
+                  unsigned char *src, 
+                  unsigned char *dst, 
+                  long           samples)
 {
   long n = samples;
 
@@ -295,11 +315,13 @@ conv_rgba8_rgbaF (unsigned char *src, unsigned char *dst, long samples)
       dst             += 4;
       src             += 1;
     }
-  return samples;
 }
 
-static INLINE long
-conv_rgb8_rgbaF (unsigned char *src, unsigned char *dst, long samples)
+static void
+conv_rgb8_rgbaF (const Babl    *conversion,
+                 unsigned char *src, 
+                 unsigned char *dst, 
+                 long           samples)
 {
   long n = samples;
 
@@ -320,11 +342,11 @@ conv_rgb8_rgbaF (unsigned char *src, unsigned char *dst, long samples)
       (*(float *) dst) = 1.0;
       dst             += 4;
     }
-  return samples;
 }
 
-static long
-conv_rgbAF_rgb8 (unsigned char *srcc,
+static void
+conv_rgbAF_rgb8 (const Babl    *conversion,
+                 unsigned char *srcc,
                  unsigned char *dstc,
                  long           samples)
 {
@@ -335,15 +357,15 @@ conv_rgbAF_rgb8 (unsigned char *srcc,
   while (n--)
     {
       float alpha = src[3];
-      if (alpha < BABL_ALPHA_THRESHOLD)
+      if (alpha == 0.0f)
         {
-          dst[0] = 0;
-          dst[1] = 0;
-          dst[2] = 0;
+          dst[0] = 0.0f;
+          dst[1] = 0.0f;
+          dst[2] = 0.0f;
         }
       else
         {
-          float alpha_recip = 1.0 / alpha;
+          float alpha_recip = 1.0f / alpha;
           dst[0] = table_F_8g[gggl_float_to_index16 (src[0] * alpha_recip)];
           dst[1] = table_F_8g[gggl_float_to_index16 (src[1] * alpha_recip)];
           dst[2] = table_F_8g[gggl_float_to_index16 (src[2] * alpha_recip)];
@@ -351,11 +373,11 @@ conv_rgbAF_rgb8 (unsigned char *srcc,
       src += 4;
       dst += 3;
     }
-  return samples;
 }
 
-static long
-conv_bgrA8_rgba8 (unsigned char *srcc,
+static void
+conv_bgrA8_rgba8 (const Babl    *conversion,
+                  unsigned char *srcc,
                   unsigned char *dstc,
                   long           samples)
 {
@@ -373,12 +395,11 @@ conv_bgrA8_rgba8 (unsigned char *srcc,
       src   += 4;
       dst   += 4;
     }
-  return samples;
 }
 
-
-static long
-conv_rgbaF_rgbAF (unsigned char *srcc,
+static void
+conv_rgbaF_rgbAF (const Babl    *conversion,
+                  unsigned char *srcc,
                   unsigned char *dstc,
                   long           samples)
 {
@@ -389,19 +410,19 @@ conv_rgbaF_rgbAF (unsigned char *srcc,
   while (n--)
     {
       float alpha = src[3];
-      dst[0] = src[0] * alpha;
-      dst[1] = src[1] * alpha;
-      dst[2] = src[2] * alpha;
+      float used_alpha = babl_epsilon_for_zero_float (alpha);
+      dst[0] = src[0] * used_alpha;
+      dst[1] = src[1] * used_alpha;
+      dst[2] = src[2] * used_alpha;
       dst[3] = alpha;
       src   += 4;
       dst   += 4;
     }
-  return samples;
 }
 
-
-static long
-conv_rgbAF_rgbaF (unsigned char *srcc,
+static void
+conv_rgbAF_rgbaF (const Babl    *conversion,
+                  unsigned char *srcc,
                   unsigned char *dstc,
                   long           samples)
 {
@@ -412,25 +433,23 @@ conv_rgbAF_rgbaF (unsigned char *srcc,
   while (n--)
     {
       float alpha = src[3];
-      float recip;
-      if (alpha < BABL_ALPHA_THRESHOLD)
-        recip = 0.0;
-      else
-        recip = 1.0/alpha;
+      float used_alpha = babl_epsilon_for_zero_float (alpha);
+      float recip = 1.0f/used_alpha;
+
       dst[0] = src[0] * recip;
       dst[1] = src[1] * recip;
       dst[2] = src[2] * recip;
+
       dst[3] = alpha;
       src   += 4;
       dst   += 4;
     }
-  return samples;
 }
 
 
-
-static long
-conv_rgbAF_lrgba8 (unsigned char *srcc,
+static void
+conv_rgbAF_lrgba8 (const Babl    *conversion,
+                   unsigned char *srcc,
                    unsigned char *dstc,
                    long           samples)
 {
@@ -441,25 +460,40 @@ conv_rgbAF_lrgba8 (unsigned char *srcc,
   while (n--)
     {
       float alpha = src[3];
-      float recip = (1.0/alpha);
-      if (alpha < BABL_ALPHA_THRESHOLD)
-        {
-          dst[0] = dst[1] = dst[2] = dst[3] = 0;
-        }
-      else
-        {
-          dst[0] = table_F_8[gggl_float_to_index16 (src[0] * recip)];
-          dst[1] = table_F_8[gggl_float_to_index16 (src[1] * recip)];
-          dst[2] = table_F_8[gggl_float_to_index16 (src[2] * recip)];
-          dst[3] = table_F_8[gggl_float_to_index16 (alpha)];
-        }
+      float used_alpha = babl_epsilon_for_zero_float (alpha);
+      float recip = (1.0f/used_alpha);
+
+      dst[0] = table_F_8[gggl_float_to_index16 (src[0] * recip)];
+      dst[1] = table_F_8[gggl_float_to_index16 (src[1] * recip)];
+      dst[2] = table_F_8[gggl_float_to_index16 (src[2] * recip)];
+      dst[3] = table_F_8[gggl_float_to_index16 (alpha)];
       src   += 4;
       dst   += 4;
     }
-  return samples;
+}
+
+static void
+conv_rgba8_rgb8 (const Babl    *conversion,
+                 unsigned char *src, 
+                 unsigned char *dst, 
+                 long           samples)
+{
+  long n = samples - 1;
+
+  while (n--)
+    {
+      *(unsigned int *) dst = (*(unsigned int *) src);
+      src   += 4;
+      dst   += 3;
+    }
+  dst[0] = src[0];
+  dst[1] = src[1];
+  dst[2] = src[2];
 }
 
 #define conv_rgb8_rgbAF    conv_rgb8_rgbaF
+#define conv_gamma_rgbaF_gamma_rgbAF   conv_rgbaF_rgbAF
+#define conv_gamma_rgbAF_gamma_rgbaF   conv_rgbAF_rgbaF
 
 int init (void);
 
@@ -482,6 +516,23 @@ init (void)
     babl_component ("Ba"),
     babl_component ("A"),
     NULL);
+    
+  const Babl *gamma_rgbaF = babl_format_new (
+    babl_model ("R'G'B'A"),
+    babl_type ("float"),
+    babl_component ("R'"),
+    babl_component ("G'"),
+    babl_component ("B'"),
+    babl_component ("A"),
+    NULL);
+  const Babl *gamma_rgbAF = babl_format_new (
+    babl_model ("R'aG'aB'aA"),
+    babl_type ("float"),
+    babl_component ("R'a"),
+    babl_component ("G'a"),
+    babl_component ("B'a"),
+    babl_component ("A"),
+    NULL);
 
   const Babl *lrgba8 = babl_format_new (
     babl_model ("RGBA"),
@@ -524,13 +575,16 @@ init (void)
 
   o (rgbaF, rgbAF);
   o (rgbAF, rgbaF);
+  o (gamma_rgbaF, gamma_rgbAF);
+  o (gamma_rgbAF, gamma_rgbaF);
   o (rgbAF, lrgba8);
-  o (rgb8, rgbaF);
-  o (rgb8, rgbAF);
+  o (rgb8,  rgbaF);
+  o (rgb8,  rgbAF);
   o (rgba8, rgbaF);
   o (rgbaF, rgb8);
   o (rgbAF, rgb8);
   o (bgrA8, rgba8);
+  o (rgba8, rgb8);
 
   return 0;
 }
